from .junitparser import (
    JUnitXmlError,
    Attr,
    Element,
    JUnitXml,
    TestSuite,
    Property,
    Skipped,
    Failure,
    Error,
    TestCase,
    Properties,
    IntAttr,
    FloatAttr,
)

version = "3.2.0"
